// Data format class for CoNLL-X format

#ifndef CONLL_FORMAT_HPP
#define CONLL_FORMAT_HPP

#include "Stack.hpp"
#include <boost/lexical_cast.hpp>

class CoNLLFormat {
private:
  size_t line_no;

public:
  enum FeatureIndex {
    SURFACE, BASE, BASE_POS, POS, FEAT
  };

public:
  CoNLLFormat() : line_no(0) {}

  size_t getLineNo() const { return line_no; }
  void clearLineNo() { line_no = 0; }

protected:
  // map integer to string
  std::string mapLength(int len) {
    if (len < 0) return "-1";
    if (len == 0) return "0";
    if (len <= 1) return "1";
    if (len <= 2) return "2";
    if (len <= 4) return "4";
    if (len <= 8) return "8";
    if (len <= 16) return "16";
    if (len <= 32) return "32";
    return "33+";
  }
  std::string mapLength(WordID left, WordID right) {
    int len = right - left;
    return mapLength(len);
  }

  std::vector<int> map2(int a, int b) {
    std::vector<int> x(2);
    x[0] = a; x[1] = b;
    return x;
  }
  std::vector<int> map3(int a, int b, int c) {
    std::vector<int> x(3);
    x[0] = a; x[1] = b; x[2] = c;
    return x;
  }
  std::vector<int> map4(int a, int b, int c, int d) {
    std::vector<int> x(4);
    x[0] = a; x[1] = b; x[2] = c; x[3] = d;
    return x;
  }
  std::vector<int> map5(int a, int b, int c, int d, int e) {
    std::vector<int> x(5);
    x[0] = a; x[1] = b; x[2] = c; x[3] = d; x[4] = e;
    return x;
  }

public:
  // extract features
  template <typename Functor>
  void makeFeature(const std::vector<Word>& words,
                   const std::vector<Dependency>&,
                   const Stack& stack,
                   WordID next_word,
                   StringIndexer* token_indexer,
                   Functor func) {
    const StackElem& left = stack.leftElem();
    const StackElem& right = stack.rightElem();
    static int BOS = token_indexer->mapint("BOS");
    static int EOS = token_indexer->mapint("EOS");
    static int DUMMY = token_indexer->mapint("DUMMY");
    //cerr << "left=" << left << ", right=" << right << ", size=" << words.size() << endl;
    // surface forms
    int lsurf = token_indexer->mapint(words[left.word_id][SURFACE]);
    int rsurf = token_indexer->mapint(words[right.word_id][SURFACE]);
    int nsurf = (size_t)next_word < words.size() ? token_indexer->mapint(words[next_word][SURFACE]) : EOS;
    //int n1surf = (size_t)next_word + 1 < words.size() ? token_indexer->mapint(words[next_word + 1][SURFACE]) : EOS;
    //int s1surf = stack.size() > 2 ? token_indexer->mapint(words[stack[stack.size() - 3].word_id][SURFACE]) : BOS;
    // POSs
    int lpos = token_indexer->mapint(words[left.word_id][POS]);
    int rpos = token_indexer->mapint(words[right.word_id][POS]);
    int npos = (size_t)next_word < words.size() ? token_indexer->mapint(words[next_word][POS]) : EOS;
    int n1pos = (size_t)next_word + 1 < words.size() ? token_indexer->mapint(words[next_word + 1][POS]) : EOS;
    //int n2pos = (size_t)next_word + 2 < words.size() ? token_indexer->mapint(words[next_word + 2][POS]) : EOS;
    int s1pos = stack.size() > 2 ? token_indexer->mapint(words[stack[stack.size() - 3].word_id][POS]) : BOS;
    // left/right POSs
    //int l1surf = left.word_id > 0 ? token_indexer->mapint(words[left.word_id - 1][SURFACE]) : BOS;
    //int r1surf = (size_t)right.word_id + 1 < words.size() ? token_indexer->mapint(words[right.word_id + 1][SURFACE]) : EOS;
    //int l1pos = left.word_id > 0 ? token_indexer->mapint(words[left.word_id - 1][POS]) : BOS;
    //int r1pos = (size_t)right.word_id + 1 < words.size() ? token_indexer->mapint(words[right.word_id + 1][POS]) : EOS;
    // Base POSs
    //int lbasepos = token_indexer->mapint(words[left.word_id][BASE_POS]);
    //int rbasepos = token_indexer->mapint(words[right.word_id][BASE_POS]);
    //int nbasepos = (size_t)next_word < words.size() ? token_indexer->mapint(words[next_word][BASE_POS]) : EOS;
    // distance
    //int lrdist = token_indexer->mapint(mapLength(left.word_id, right.word_id));
    //int rndist = token_indexer->mapint(mapLength(right.word_id, next_word));
    // number of children
    //int llchld = token_indexer->mapint(mapLength(left.left.size()));
    //int lrchld = token_indexer->mapint(mapLength(left.right.size()));
    //int lchld = token_indexer->mapint(mapLength(left.left.size() + left.right.size()));
    //int rlchld = token_indexer->mapint(mapLength(right.left.size()));
    //int rrchld = token_indexer->mapint(mapLength(right.right.size()));
    //int rchld = token_indexer->mapint(mapLength(right.left.size() + right.right.size()));
    // tree size
    //int lsize = token_indexer->mapint(mapLength(left.size));
    //int rsize = token_indexer->mapint(mapLength(right.size));
    // left/right-most words
    //int llsurf = left.left.empty() ? DUMMY : token_indexer->mapint(words[left.left.front()][SURFACE]);
    //int rlsurf = right.left.empty() ? DUMMY : token_indexer->mapint(words[right.left.front()][SURFACE]);
    int llpos = left.left.empty() ? DUMMY : token_indexer->mapint(words[left.left.front().word_id][POS]);
    int rlpos = right.left.empty() ? DUMMY : token_indexer->mapint(words[right.left.front().word_id][POS]);
    //int lldep = left.left.empty() ? DUMMY : token_indexer->mapint(dep_tree[left.left.front()].label);
    //int rldep = right.left.empty() ? DUMMY : token_indexer->mapint(dep_tree[right.left.front()].label);
    //int lrsurf = left.right.empty() ? DUMMY : token_indexer->mapint(words[left.right.back()][SURFACE]);
    //int rrsurf = right.right.empty() ? DUMMY : token_indexer->mapint(words[right.right.back()][SURFACE]);
    int lrpos = left.right.empty() ? DUMMY : token_indexer->mapint(words[left.right.back().word_id][POS]);
    int rrpos = right.right.empty() ? DUMMY : token_indexer->mapint(words[right.right.back().word_id][POS]);
    //int lrdep = left.right.empty() ? DUMMY : token_indexer->mapint(dep_tree[left.right.back()].label);
    //int rrdep = right.right.empty() ? DUMMY : token_indexer->mapint(dep_tree[right.right.back()].label);
    // previous action
//     int prevaction;
//     if (stack.prevActions().back() == ShiftReduceParserBase::SHIFT) {
//       prevaction = token_indexer->mapint("_pa_S");
//     } else if (stack.prevActions().back() == ShiftReduceParserBase::REDUCE_LEFT) {
//       prevaction = token_indexer->mapint("_pa_L" + dep_tree[right.right.back()].label);
//     } else {
//       prevaction = token_indexer->mapint("_pa_R" + dep_tree[right.left.front()].label);
//     }

    int template_id = 0;
    func(map2(lsurf, ++template_id), 1.0);
    func(map2(rsurf, ++template_id), 1.0);
    func(map2(nsurf, ++template_id), 1.0);
    func(map2(lpos, ++template_id), 1.0);
    func(map2(rpos, ++template_id), 1.0);
    func(map2(npos, ++template_id), 1.0);
    func(map3(lsurf, lpos, ++template_id), 1.0);
    func(map3(rsurf, rpos, ++template_id), 1.0);
    func(map3(nsurf, npos, ++template_id), 1.0);

    func(map3(lsurf, rsurf, ++template_id), 1.0);
    func(map3(lpos, rpos, ++template_id), 1.0);
    func(map3(rpos, npos, ++template_id), 1.0);
    func(map4(lsurf, lpos, rpos, ++template_id), 1.0);
    func(map4(rsurf, rpos, lpos, ++template_id), 1.0);
    func(map4(lsurf, lpos, rsurf, ++template_id), 1.0);
    func(map4(rsurf, rpos, lsurf, ++template_id), 1.0);
    func(map5(lsurf, lpos, rsurf, rpos, ++template_id), 1.0);

    func(map4(rpos, npos, n1pos, ++template_id), 1.0);
    func(map4(lpos, rpos, npos, ++template_id), 1.0);
    func(map4(rsurf, npos, n1pos, ++template_id), 1.0);
    func(map4(lpos, rsurf, npos, ++template_id), 1.0);

    func(map4(lpos, rpos, llpos, ++template_id), 1.0);
    func(map4(lpos, rpos, lrpos, ++template_id), 1.0);
    func(map4(lpos, rpos, rrpos, ++template_id), 1.0);
    func(map4(lpos, rpos, rlpos, ++template_id), 1.0);
    func(map4(lpos, rsurf, lrpos, ++template_id), 1.0);
    func(map4(lpos, rsurf, rlpos, ++template_id), 1.0);

    func(map4(s1pos, lpos, rpos, ++template_id), 1.0);
  }

  // read CoNLL-X format
  void readDependencies(std::istream& is, std::vector<Word>& words, std::vector<Dependency>& dep_tree) {
    std::string line, str;
    while (getline(is, line)) {
      ++line_no;
      std::istringstream iss(line);
      std::vector<std::string> tokens;
      while (iss >> str) {
        tokens.push_back(str);
      }
      if (tokens.empty()) break;
      if (tokens.size() != 10
          || (int)words.size() != boost::lexical_cast<WordID>(tokens[0]) - 1) {
        throw Exception("Wrong format for CoNLL: " + line);
      }
      words.push_back(Word(tokens.begin() + 1, tokens.begin() + 6));
      dep_tree.push_back(Dependency(boost::lexical_cast<WordID>(tokens[6]) - 1,
                                    //tokens[7]));
                                    "DEP"));
    }
  }

  // output parse result in CoNLL-X format
  void writeDependencies(std::ostream& output, const std::vector<Word>& words, const std::vector<Dependency>& dep_tree) {
    // output dep tree
    for (std::vector<Word>::size_type i = 0; i < words.size(); ++i) {
      std::string label = dep_tree[i].label.empty() ? "ROOT" : dep_tree[i].label;
      output << i + 1 << "\t";
      output << words[i][SURFACE] << "\t";
      output << words[i][BASE] << "\t";
      output << words[i][BASE_POS] << "\t";
      output << words[i][POS] << "\t";
      output << words[i][FEAT] << "\t";
      output << dep_tree[i].head + 1 << "\t";
      output << label << "\t_\t_\n";
    }
    output << "\n";
  }
};

#endif // CONLL_FORMAT_HPP

