#ifndef LOOK_AHEAD_PARSER_HPP
#define LOOK_AHEAD_PARSER_HPP

#include "Stack.hpp"
#include "StringIndexer.hpp"
#include "WeightMap.hpp"
#include "FeatureVec.hpp"
#include "Random.hpp"
#include <vector>
#include <string>
#include <deque>
#include <set>
#include <boost/lexical_cast.hpp>
#include <cassert>

class TrainingInstance {
public:
  std::vector<Word> words;
  std::vector<Dependency> dep_tree;
  std::vector<LabelID> actions;

  TrainingInstance(const std::vector<Word>& w, const std::vector<Dependency>& d, const std::vector<LabelID>& a) : words(w), dep_tree(d), actions(a) {}
};

template <typename DataFormat>
class LookAheadParser {
public:
  typedef double Number;
  typedef int LabelID;
  typedef std::vector<int> FeatureName;
  typedef BinaryFeature<FeatureName> Feature;
  typedef FeatureVec<Feature, Number> FeatureVector;
  typedef LabelID Target;
  typedef std::vector<FeatureName> History;
  typedef std::deque<TrainingInstance> Corpus;

  static const bool DO_AVERAGING = true;

  class FeatureHashFun {
  public:
    size_t operator()(const std::pair<Target, FeatureName>& f) const {
      size_t x = f.first;
      for (FeatureName::const_iterator it = f.second.begin();
           it != f.second.end();
           ++it) {
        x = x * 100 + *it;
      }
      return x;
    }
  };

  class FeatureEqualFun {
  public:
    bool operator()(const std::pair<Target, FeatureName>& x,
                    const std::pair<Target, FeatureName>& y) const {
      if (x.first != y.first) return false;
      if (x.second.size() != y.second.size()) return false;
      FeatureName::const_iterator x_it = x.second.begin();
      FeatureName::const_iterator y_it = y.second.begin();
      for (; x_it != x.second.end(); ++x_it, ++y_it) {
        if (y_it == y.second.end()) return false;
        if (*x_it != *y_it) return false;
      }
      if (y_it != y.second.end()) return false;
      return true;
    }
  };

  typedef WeightHash<std::pair<Target, FeatureName>, Number, FeatureHashFun, FeatureEqualFun > WeightMap;

protected:
  DataFormat* data_format;
  StringIndexer* rel_indexer;
  StringIndexer* token_indexer;
  std::set<LabelID> label_set;

  WeightMap weight;
  WeightMap weight2;
  int search_depth;
  Number perceptron_margin;

  int num_decode;
  int num_update;
  bool is_converged;

  Random rand_generator;

  LabelID encodeAction(Action action, int deplabel) {
    if (action == SHIFT) return 0;
    return deplabel * 2 + action;
  }
  std::pair<Action, int> decodeAction(LabelID label) {
    if (label == 0) return std::make_pair(SHIFT, 0);
    int rel = (label - 1) / 2;
    if (label % 2 == 0) return std::make_pair(REDUCE_RIGHT, rel);
    return std::make_pair(REDUCE_LEFT, rel);
  }

public:
  LookAheadParser(DataFormat* f, StringIndexer* li, StringIndexer* ti, int d, Number m = 0.0)
    : data_format(f), rel_indexer(li), token_indexer(ti), weight(0), weight2(0), search_depth(d), perceptron_margin(m), num_decode(0), num_update(0), is_converged(false) {
    label_set.insert(encodeAction(SHIFT, 0));
  }
  virtual ~LookAheadParser() {}

  size_t numRelations() const { return rel_indexer->size(); }
  size_t numLabels() const { return label_set.size(); }
  size_t numFeatures() const { return weight.size(); }

public:
  const WeightMap& getWeight() const { return weight; }
  WeightMap& getWeight() { return weight; }

  Number score(Target action, const History& history) {
    Number score = 0.0;
    for (History::const_iterator it = history.begin();
         it != history.end();
         ++it) {
      score += weight.weight(std::make_pair(action, *it));
    }
    return score;
  }

protected:
  class PushHistory {
  public:
    History* history;
    PushHistory(History* h) : history(h) {}
    void operator()(const FeatureName& f, Number) {
      history->push_back(f);
    }
  };

  // run look-ahead search
  Number lookaheadSearch(const std::vector<Word>& words, const std::vector<Dependency>& dep_tree, Stack& stack, WordID next_word, int depth, std::vector<LabelID>& action_list, LabelID correct_action = 0, Number margin = 0.0) {
    if (depth == 0) return 0.0;
    if ((size_t)next_word == words.size() && stack.size() == 1) return 0.0;  // parsing completed
    if (stack.size() < 2) {
      // only SHIFT allowed
      std::vector<LabelID> actions;
      stack.shift(next_word);
      Number score = lookaheadSearch(words, dep_tree, stack, next_word + 1, depth - 1, actions);
      stack.rewindShift();
      action_list.resize(actions.size() + 1);
      action_list[0] = 0;
      copy(actions.begin(), actions.end(), action_list.begin() + 1);
      return score;
    } else {
      Number max_score = -std::numeric_limits<Number>::max();
      History history;
      data_format->makeFeature(words, dep_tree, stack, next_word, token_indexer,
                               PushHistory(&history));
      for (std::set<LabelID>::const_iterator label_it = label_set.begin();
           label_it != label_set.end();
           ++label_it) {
        Number score = this->score(*label_it, history);
        std::vector<LabelID> actions;
        if (*label_it == 0) {
          // SHIFT
          if ((size_t)next_word < words.size()) {
            stack.shift(next_word);
            score += lookaheadSearch(words, dep_tree, stack, next_word + 1, depth - 1, actions);
            stack.rewindShift();
          }
        } else {
          if (decodeAction(*label_it).first == REDUCE_LEFT) {
            // REDUCE LEFT
            stack.reduceLeft();
            score += lookaheadSearch(words, dep_tree, stack, next_word, depth - 1, actions);
            stack.rewindReduceLeft();
          } else {
            // REDUCE RIGHT
            stack.reduceRight();
            score += lookaheadSearch(words, dep_tree, stack, next_word, depth - 1, actions);
            stack.rewindReduceRight();
          }
        }
        //std::cerr << *label_it << ": " << score << std::endl;
        if (margin > 0.0 && *label_it != correct_action) {
          score += margin;  // add margin
        }
        if (score > max_score) {
          action_list.resize(actions.size() + 1);
          action_list[0] = *label_it;
          copy(actions.begin(), actions.end(), action_list.begin() + 1);
          max_score = score;
        }
      }
      return max_score;
    }
  }

public:
  bool isConverged() { return is_converged; }
  std::string showStatistics() {
    std::ostringstream os;
    os << num_decode << " decoding, " << num_update << " updates";
    return os.str();
  }

  // training for a sentence
  void trainSentence(const std::vector<Word>& words, const std::vector<Dependency>& dep_tree, const std::vector<LabelID>& correct_action_list) {
    Stack stack;
    WordID next_word = 0;
    for (std::vector<LabelID>::const_iterator action_it = correct_action_list.begin();
         action_it != correct_action_list.end();
         ++action_it) {
      if (stack.size() < 2) {
        assert(*action_it == 0);  // should be SHIFT
        stack.shift(next_word);
        ++next_word;
      } else {
        ++num_decode;
        std::vector<LabelID> action_list;
        lookaheadSearch(words, dep_tree, stack, next_word, search_depth, action_list, *action_it, perceptron_margin);
        if (action_list.front() != *action_it) {
          ++num_update;
          updateWeights(words, dep_tree, stack, next_word, action_list.begin(), action_list.end(), -1);
          updateWeights(words, dep_tree, stack, next_word, action_it, min(action_it + search_depth, correct_action_list.end()), 1);
        }
        if (*action_it == 0) {
          // SHIFT
          stack.shift(next_word);
          ++next_word;
        } else {
          if (decodeAction(*action_it).first == REDUCE_LEFT) {
            // REDUCE LEFT
            stack.reduceLeft();
          } else {
            stack.reduceRight();
          }
        }
      }
    }
  }

  class UpdateWeights {
  public:
    WeightMap* weight;
    WeightMap* weight2;
    LabelID action;
    Number diff;
    int num_decode;
    UpdateWeights(WeightMap* w, WeightMap* w2, LabelID a, Number d, int n) : weight(w), weight2(w2), action(a), diff(d), num_decode(n) {}
    void operator()(const FeatureName& f, Number) {
      std::pair<LabelID, FeatureName> p(action, f);
//       if (weight->find(p) == weight->end()) {
//         (*weight)[p] = diff;
//         (*weight2)[p] = diff * num_decode;
//         //(*weight2)[p] = diff;
//       } else {
        (*weight)[p] += diff;
        (*weight2)[p] += diff * num_decode;
//       }
    }
  };

  // perceptron-like update
  void updateWeights(const std::vector<Word>& words,
                     const std::vector<Dependency>& dep_tree,
                     Stack& stack,
                     WordID next_word,
                     std::vector<LabelID>::const_iterator action_it,
                     std::vector<LabelID>::const_iterator action_end,
                     Number diff) {
    if (action_it == action_end) return;
    if (stack.size() < 2) {
      assert(*action_it == 0);  // should be SHIFT
      stack.shift(next_word);
      updateWeights(words, dep_tree, stack, next_word + 1,
                    action_it + 1, action_end, diff);
      stack.rewindShift();
    } else {
      data_format->makeFeature(words, dep_tree, stack, next_word, token_indexer,
                               UpdateWeights(&weight, &weight2, *action_it, diff, num_decode));
      if (*action_it == 0) {
        stack.shift(next_word);
        updateWeights(words, dep_tree, stack, next_word + 1,
                      action_it + 1, action_end, diff);
        stack.rewindShift();
      } else if (decodeAction(*action_it).first == REDUCE_LEFT) {
        stack.reduceLeft();
        updateWeights(words, dep_tree, stack, next_word,
                      action_it + 1, action_end, diff);
        stack.rewindReduceLeft();
      } else {
        stack.reduceRight();
        updateWeights(words, dep_tree, stack, next_word,
                      action_it + 1, action_end, diff);
        stack.rewindReduceRight();
      }
    }
  }

  //////////////////////////////////////////////////////////////////////

protected:
  // run shift-reduce parsing using oracle
  std::vector<LabelID> oracleParse(const std::vector<Word>& words, const std::vector<Dependency>& dep_tree) {
    WordID next_word = 0;
    std::vector<LabelID> action_list;
    Stack stack;
    while (true) {
      if ((size_t)next_word == words.size() && stack.size() == 1) break;  // parsing completed
      LabelID action = 0;
      if (stack.size() >= 2) {
        const StackElem& left = stack.leftElem();
        const StackElem& right = stack.rightElem();
        WordID nonhead_id = 0;
        if (dep_tree[right.word_id].head == left.word_id) {
          // reduce left
          // check whether right words do not modify this word
          std::vector<Dependency>::const_iterator it = dep_tree.begin() + next_word;
          for (; it != dep_tree.end(); ++it) {
            if (it->head == right.word_id) break;
          }
          if (it == dep_tree.end()) {  // if not, SHIFT
            int rel = rel_indexer->getint(dep_tree[right.word_id].label, true);
            action = encodeAction(REDUCE_LEFT, rel);
            label_set.insert(action);
            nonhead_id = right.word_id;
          }
        } else if (dep_tree[left.word_id].head == right.word_id) {
          // reduce right
          int rel = rel_indexer->getint(dep_tree[left.word_id].label, true);
          action = encodeAction(REDUCE_RIGHT, rel);
          label_set.insert(action);
          nonhead_id = left.word_id;
        } else if ((size_t)next_word == words.size()) {
          // parsing not completed -- maybe non-projective
          std::cerr << "Maybe nonprojective at line " << data_format->getLineNo() << std::endl;
          //break;
          exit(1);
        }
      }
      action_list.push_back(action);
      if (action == 0) {
        stack.shift(next_word);
        ++next_word;
      } else if (decodeAction(action).first == REDUCE_LEFT) {
        stack.reduceLeft();
      } else {
        stack.reduceRight();
      }
    }
    return action_list;
  }

public:
  // read training data from a corpus
  void readTrainingData(std::istream& is, Corpus& training_data) {
    data_format->clearLineNo();
    size_t sentence_no = 0;
    while (is) {
      std::vector<Word> words;
      std::vector<Dependency> dep_tree;
      data_format->readDependencies(is, words, dep_tree);
      if (words.empty()) continue;
      training_data.push_back(TrainingInstance(words, dep_tree, this->oracleParse(words, dep_tree)));
      ++sentence_no;
      if (sentence_no % 1000 == 0) {
        std::cerr << "  " << sentence_no << " sentences input" << std::endl;
      }
    }
    std::cerr << "  " << sentence_no << " sentences input" << std::endl;
  }

  // train a model
  void initialize() {
  }

  void iterate(const Corpus& training_data) {
    std::vector<int> index(training_data.size());
    for (int i = 0; i < (int)index.size(); ++i) {
      index[i] = i;
    }
    random_shuffle(index.begin(), index.end(), rand_generator);
    int prev_num_update = num_update;
    for (std::vector<int>::const_iterator it = index.begin();
         it != index.end();
         ++it) {
      trainSentence(training_data[*it].words,
                    training_data[*it].dep_tree,
                    training_data[*it].actions);
    }
//     for (Corpus::const_iterator sentence_it = training_data.begin();
//          sentence_it != training_data.end();
//          ++sentence_it) {
//       //std::cerr << sentence_it - training_data.begin() << std::endl;
//       trainSentence(sentence_it->words, sentence_it->dep_tree, sentence_it->actions);
//     }
    if (prev_num_update == num_update) is_converged = true;
  }

  void finalize() {
  }

  //////////////////////////////////////////////////////////////////////

  void parse(const std::vector<Word>& words, std::vector<Dependency>& dep_tree) {
    dep_tree.resize(words.size());
    for (std::vector<Dependency>::iterator it = dep_tree.begin();
         it != dep_tree.end();
         ++it) {
      it->head = -1;
    }
    Stack stack;
    WordID next_word = 0;
    while (true) {
      //std::cerr << next_word << " " << stack.size() << std::endl;
      if ((size_t)next_word == words.size() && stack.size() == 1) break;  // parsing completed
      if (stack.size() >= 2) {
        std::vector<LabelID> action_list;
        lookaheadSearch(words, dep_tree, stack, next_word, search_depth, action_list);
        LabelID action = action_list.front();
        if (action == 0) {
          // SHIFT
          if ((size_t)next_word == words.size()) break;  // parsing not completed
          stack.shift(next_word);
          ++next_word;
        } else {
          const StackElem& left = stack.leftElem();
          const StackElem& right = stack.rightElem();
          if (decodeAction(action).first == REDUCE_LEFT) {
            // REDUCE LEFT
            dep_tree[right.word_id].head = left.word_id;
            dep_tree[right.word_id].label = rel_indexer->getString(decodeAction(action).second);
            stack.reduceLeft();
          } else {
            dep_tree[left.word_id].head = right.word_id;
            dep_tree[left.word_id].label = rel_indexer->getString(decodeAction(action).second);
            stack.reduceRight();
          }
        }
      } else {
        if ((size_t)next_word == words.size()) break;  // parsing not completed
        stack.shift(next_word);
        ++next_word;
      }
    }
    // output root dependency
    for (std::vector<StackElem>::const_iterator it = stack.begin();
         it != stack.end();
         ++it) {
      dep_tree[it->word_id].head = -1;
      dep_tree[it->word_id].label = "";
    }
  }

  void parseSentence(std::istream& in, std::ostream& out) {
    // read sentence
    std::vector<Word> words;
    std::vector<Dependency> dep_tree;
    data_format->readDependencies(in, words, dep_tree);
    // parse
    this->parse(words, dep_tree);
    // write dependencies
    data_format->writeDependencies(out, words, dep_tree);
  }

  //////////////////////////////////////////////////////////////////////

  void writeWeight(std::ostream& os) {
    for (typename WeightMap::const_iterator it = weight.begin();
         it != weight.end();
         ++it) {
      os << it->first.first;
      for (typename FeatureName::const_iterator it2 = it->first.second.begin();
           it2 != it->first.second.end();
           ++it2) {
        os << ' ' << *it2;
      }
      Number score = it->second;
      if (DO_AVERAGING) {
        score -= weight2[it->first] / static_cast<Number>(num_decode);
      }
      os << '\t' << score << '\n';
    }
  }

  void readWeight(std::istream& is) {
    std::string line, token;
    std::vector<std::string> tokens;
    std::vector<int> feature;
    while (getline(is, line)) {
      std::istringstream istr(line);
      tokens.clear();
      while (istr >> token) {
        tokens.push_back(token);
      }
      Number w = boost::lexical_cast<Number>(tokens.back());
      LabelID label = boost::lexical_cast<LabelID>(tokens.front());
      label_set.insert(label);
      feature.resize(tokens.size() - 2);
      for (size_t i = 0; i < tokens.size() - 2; ++i) {
        feature[i] = boost::lexical_cast<int>(tokens[i + 1]);
      }
      weight[std::make_pair(label, feature)] = w;
    }
  }
};

#endif // LOOK_AHEAD_PARSER_HPP
