#include "LookAheadParser.hpp"
#include "CoNLLFormat.hpp"
#include "FileIO.hpp"
#include "Timer.hpp"
#include <boost/lexical_cast.hpp>

using namespace std;

void printUsage(ostream &out) {
  out << "Usage: TestLookAheadParser -e TEST_FILE [options]" << endl;
}

int main(int argc, char** argv) {
  const char *opt_string="e:l:v:w:m:d:";
  string test(""), label("label.dat"), vocab("vocab.dat"), feat("feat.dat"), format("conll");
  int search_depth = 1;
  int c;
  while ((c = getopt(argc, argv, opt_string)) != -1) {
    switch (c) {
    case 'e': test = optarg; break;
    case 'l': label = optarg; break;
    case 'v': vocab = optarg; break;
    case 'w': feat = optarg; break;
    case 'm': format = optarg; break;
    case 'd': search_depth = boost::lexical_cast<int>(optarg); break;
    default: cerr << "Unknown option: " << c << endl; printUsage(cerr); exit(1);
    }
  }

  if (test == "") {
    printUsage(cerr);
    exit(1);
  }

  Timer timer;
  timer.start();

  // create parser
  CoNLLFormat data_format;
  StringIndexer l_index;
  StringIndexer t_index;
  LookAheadParser<CoNLLFormat> parser(&data_format, &l_index, &t_index, search_depth);
  LookAheadParser<CoNLLFormat>::Corpus test_data;

  try {
    istream* label_file = openInputStream(label);
    istream* vocab_file = openInputStream(vocab);
    istream* feat_file = openInputStream(feat);

    if (label_file->good()) {
      l_index.read(*label_file);
    } else {
      cerr << "can't read input from file '" << label << "'" << endl;
      exit(1);
    }
    if(vocab_file->good()){
      t_index.read(*vocab_file);
    } else {
      cerr << "can't read input from file '" << vocab << "'" << endl;
      exit(1);
    }
    if(feat_file->good()){
      parser.readWeight(*feat_file);
    } else {
      cerr << "can't read input from file '" << feat << "'" << endl;
      exit(1);
    }

    delete feat_file;
    delete vocab_file;
    delete label_file;

  } catch(Exception& e) {
    cerr << "error during reading parameters: " << e.message() << endl;
    exit(1);
  }

  timer.stop();
  cerr << "Data input finished in " << timer.printInterval() << endl;
  timer.clear();

  cerr << "Running the parser" << endl;

  timer.start();

  try {
    istream* test_file = openInputStream(test);
    if (test_file->good()) {
      while (*test_file) {
        parser.parseSentence(*test_file, cout);
      }
    } else {
      cerr << "Cannot open test file: " << test << endl;
      exit(1);
    }
    delete test_file;
  } catch (Exception& e) {
    cerr << "error during reading test file '" << test << "'': " << e.message() << endl;
    exit(1);
  }

  timer.stop();
  cerr << "Test finished in " << timer.printInterval() << endl;
}

