#include "LookAheadParser.hpp"
#include "CoNLLFormat.hpp"
#include "Timer.hpp"
#include "FileIO.hpp"
//#include "Perceptron.hpp"

using namespace std;

void printUsage(ostream& os) {
  os << "Usage: TrainLookAheadParser -t TRAINING_DATA [options]\n";
}

int main(int argc, char** argv) {
  const char* opt_string="t:l:v:w:i:c:d:m:o:";
  string train(""), label("label.dat"), vocab("vocab.dat"), feat("feat.dat");
  int iterations = 1;
  int cutoff = 3;
  int search_depth = 1;
  double perceptron_margin = 10.0;
  int output_interval = 0;
  int c;
  while ((c = getopt(argc, argv, opt_string)) != -1) {
    switch (c) {
    case 't': train = optarg; break;
    case 'l': label = optarg; break;
    case 'v': vocab = optarg; break;
    case 'w': feat = optarg; break;
    case 'i': iterations = boost::lexical_cast<int>(optarg); break;
    case 'c': cutoff = boost::lexical_cast<int>(optarg); break;
    case 'd': search_depth = boost::lexical_cast<int>(optarg); break;
    case 'm': perceptron_margin = boost::lexical_cast<double>(optarg); break;
    case 'o': output_interval = boost::lexical_cast<int>(optarg); break;
    default: cerr << "Unknown option: " << c << endl; printUsage(cerr); exit(1);
    }
  }

  if (train == "") {
    printUsage(cerr);
    exit(1);
  }

  Timer timer;
  timer.start();

  // create parser
  CoNLLFormat data_format;
  StringIndexer rel_indexer; // maps a string label to an index. 
  StringIndexer token_indexer(true); // maps a string token to an index. 
  LookAheadParser<CoNLLFormat> parser(&data_format, &rel_indexer, &token_indexer, search_depth, perceptron_margin);
  LookAheadParser<CoNLLFormat>::Corpus training_data;

  // read training data
  istream* train_file = openInputStream(train);
  if (!*train_file) {
    cerr << "Cannot open training file: " << train << endl;
    exit(1);
  }
  cerr << "Reading training data from " << train << "... " << endl;
  try {
    parser.readTrainingData(*train_file, training_data);
  } catch (Exception& e) {
    cerr << "error during reading corpus: " << e.message() << endl;
    exit(1);
  }
  cerr << "done." << endl;
  delete train_file;

  timer.stop();
  cerr << "Data input finished in " << timer.printInterval() << endl;
  timer.clear();

  cerr << "labels: " << parser.numLabels() << endl;
  //cerr << "features: " << parser.numFeatures() << endl;
  cerr << "instances: " << training_data.size() << endl;

  timer.start();

  parser.initialize();
  for (int i = 1; i <= iterations; ++i) {
    cerr << "iteration " << i << ": ";
    parser.iterate(training_data);
    cerr << parser.showStatistics() << endl;
    if (parser.isConverged()) break;
    if (output_interval > 0 && i % output_interval == 0) {
      cerr << "  writing feature weights..." << endl;
      ostream* feat_file = openOutputStream("iter" + boost::lexical_cast<string>(i) + "-" + feat);
      if (feat_file->good()) {
        parser.writeWeight(*feat_file);
      } else {
        cerr << "can't write data to file '" << feat << "'" << endl;
        exit(1);
      }
      delete feat_file;
    }
  }
  parser.finalize();

  cerr << "features: " << parser.numFeatures() << endl;

  timer.stop();
  cerr << "Training finished in " << timer.printInterval() << endl;
  timer.clear();

  cerr << "Writing..." << endl;

  timer.start();

  try{
    ostream* label_file = openOutputStream(label);
    ostream* vocab_file = openOutputStream(vocab);
    ostream* feat_file = openOutputStream(feat);

    if (label_file->good()) {
      rel_indexer.print(*label_file);
    } else {
      cerr << "can't write data to file '" << label << "'" << endl;
      exit(1);
    }
    if (vocab_file->good()) {
      token_indexer.print(*vocab_file);
    } else {
      cerr << "can't write data to file '" << vocab << "'" << endl;
      exit(1);
    }
    if (feat_file->good()) {
      parser.writeWeight(*feat_file);
    } else {
      cerr << "can't write data to file '" << feat << "'" << endl;
      exit(1);
    }

    delete feat_file;
    delete vocab_file;
    delete label_file;

  } catch(Exception& e) {
    cerr << "error during writing parameters: " << e.message() << endl;
    exit(1);
  }

  timer.stop();
  cerr << "Data output finished in " << timer.printInterval() << endl;
}

